# -*- coding: utf-8 -*-
import glob
import law
import luigi
import logging
import os
import re
import shutil
import six

from functools import partial
from fnmatch import fnmatch
from subprocess import PIPE

from law.util import colored

from analysis.framework import Task, HTCondorWorkflow, SpaceSeparatedListParameter
from analysis.mixins import TarballExtractionMixin
from analysis.util import createRuncard, interruptable_popen, working_directory

logger = logging.getLogger(__name__)
working_directory = partial(working_directory, logger=logger)


class Warmup(Task, TarballExtractionMixin, HTCondorWorkflow, law.LocalWorkflow):
    base_runcard = luigi.Parameter()

    default_channels = ["LO", "R", "V", "RRa", "RRb", "RV", "VV"]
    channels = SpaceSeparatedListParameter(
        default=default_channels,
        cls=luigi.Parameter,
        choices=("LO", "R", "V", "RRa", "RRb", "RV", "VV")
    )
    starting_seed = luigi.IntParameter()
    warmup_events = SpaceSeparatedListParameter(cls=luigi.IntParameter)
    warmup_iterations = SpaceSeparatedListParameter(cls=luigi.IntParameter)

    NNLOJET_RUNMODE   = 'warmup'
    NNLOJET_UNITPHASE = '! UNIT_PHASE'

    # NNLOJET frequently does NOT report errors properly (return code > 0)!
    # ==> scan stdout for these strings to check if an error occurred
    NNLOJET_ERROR_STRINGS = (
        'Error, index request',
        'Error: define a seed',
    )

    def create_branch_map(self):
        """Branch data dicts. Each branch should correspond to exactly one output."""
        return {
            i : {
                'channel': channel,
                'events': self.warmup_events[i],
                'iterations': self.warmup_iterations[i],
                'seed': i + self.starting_seed,
            }
            for i, channel in enumerate(self.channels)
        }

    def workflow_requires(self):
        return {}

    def requires(self):
        return {}

    def output(self):
        return self.remote_or_local_target(
            '{}.{}.{}.warmup.tar.gz'.format(
                self.process,
                self.branch_data['channel'],
                self.name,
            ),
        )

    def run(self):
        rundirpath = os.getcwd()
        tmpdirpath = os.path.join(rundirpath, 'tmpdir{}'.format(self.branch_data['seed']))

        logger.debug("Running job in directory: %s", rundirpath)
        outfile = os.path.basename(self.output().path)
        logfile = re.sub('.tar.gz$', '.log', outfile)
        errfile = re.sub('.tar.gz$', '.err.log', outfile)
        runfile = re.sub('.tar.gz$', '.run', outfile)

        job_ok = True  # set to False if exception is raised

        try:
            # -- create temporary directory
            if not os.path.exists(tmpdirpath):
                logger.debug("Creating temporary directory: %s", tmpdirpath)
                os.mkdir(tmpdirpath)
            elif os.path.isdir(tmpdirpath):
                logger.warning("Temporary directory already exists: %s", tmpdirpath)
            else:
                raise Exception("Could not create temporary directory: %s".format(tmpdirpath))

            # -- check and copy base run card
            logger.debug("Checking for base runcard: %s", self.base_runcard)
            if not os.path.isfile(self.base_runcard):
                raise Exception('Base runcard {} not found!'.format(self.base_runcard))
            logger.debug(
                "Copying base runcard: %s -> %s", self.base_runcard,
                os.path.join(tmpdirpath, self.base_runcard)
            )
            shutil.copy(self.base_runcard, os.path.join(tmpdirpath, self.base_runcard))

            # ensure output parent directory exists
            logger.debug("Creating output parent directory: %s", self.output().parent.path)
            self.output().parent.touch()

            # load data of all required tasks into current
            # directory (e.g. output of previous Warmup run)
            for _, input_ in six.iteritems(self.input()):
                # extract input tarball to tmpdir
                self.extract_tarball(
                    input_,
                    law.LocalDirectoryTarget(tmpdirpath),
                    formatter="tar",
                    logger=logger,
                )

            # run all subsequent command in tmpdir
            with working_directory(tmpdirpath):
                logger.debug("Out filename is: %s", outfile)
                logger.debug("Log filename is: %s", logfile)
                logger.debug("Run cardname is: %s", runfile)

                logger.debug("Perfoming in-place @@ substitution on runcard: %s", runfile)
                with open(runfile, 'w') as f:
                    runcard = createRuncard(
                        open(self.base_runcard, 'r').read(), {
                            'channel': self.branch_data['channel'],
                            'seed': self.branch_data['seed'],
                            'iterations': self.branch_data['iterations'],
                            'events': self.branch_data['events'],
                            'runmode': self.NNLOJET_RUNMODE,
                            'unit_phase': self.NNLOJET_UNITPHASE
                        }
                    )
                    f.write(runcard)

                # KR: With hyperthreading it might be necessary to request more cpus on a node than one wants to use for the calculation!
                #     --> Differentiate between request and use CPUs!
                #      os.environ['OMP_NUM_THREADS'] = self.htcondor_request_cpus
                os.environ['OMP_NUM_THREADS'] = self.htcondor_use_cpus

                logger.debug("Starting NNLOJET with runcard: %s", runfile)

                command_line_args = ['NNLOJET', '-run', runfile]
                logger.info("Running command: %s",
                    colored(' '.join(command_line_args), 'yellow'))
                code, out, error = interruptable_popen(command_line_args,
                                                       stdout=PIPE,
                                                       stderr=PIPE)
                logger.debug("Return code: %d", code)
                logger.debug("Standard output:\n%s\n", out)
                logger.debug("Standard error:\n%s\n", error)

                # code may be 0 even if an error occurs -> scan stdout for error strings
                for err_string in self.NNLOJET_ERROR_STRINGS:
                    if err_string in out:
                        code = 1
                        ###error = out
                        break

                logger.debug("Dumping standard out to file:  %s", logfile)
                with open(logfile, 'w') as f:
                    f.write(out)

                logger.debug("Dumping standard error to file:  %s", errfile)
                with open(errfile, 'w') as f:
                    f.write(error)

                if (code == 0):
                    logger.info("NNLOJET status: SUCCESS")
                    job_ok = True
                else:
                    raise RuntimeError("NNLOJET status: FAIL (exit code {})".format(code))

        # mark job as failed on error, make log entry and reraise
        except Exception as e:
            job_ok = False
            logger.error('%s raised during job execution: %s', e.__class__.__name__, e)
            raise

        # finalize job
        finally:
            logger.debug("Finalizing job...")

            with working_directory(tmpdirpath):
                logger.debug("Files in current directory:\n  %s", '\n  '.join(glob.glob('*')))

                # write out results on success, debug info on fail
                if job_ok:
                    tarfilter = lambda n: n if (
                        n.name == '.' or
                        fnmatch(n.name, '*{}*'.format(self.branch_data['channel']))
                    ) else None

                    logger.debug("Job successful. Writing output tarfile from directory: %s", os.getcwd())
                    logger.debug("Job successful. Writing output tarfile to directory: %s", self.output().path)
                    self.output().dump(os.getcwd(), mode="w:gz", formatter="tar", add_kwargs=dict(filter=tarfilter))
                else:
                    if self.htcondor_remote_storage:
                        debugfile = re.sub('.tar.gz$', '.debug.tar.gz', os.path.basename(self.output().path))
                        logger.debug("Job failed. Writing debug tarfile: %s -> %s", os.getcwd(), debugfile)
                        self.remote_target(debugfile).dump(os.getcwd(), mode="w:gz", formatter="tar")

            logger.debug("Removing temporary directory: %s", tmpdirpath)
            shutil.rmtree(tmpdirpath)
